(* mclu: Mini Cloud
 * Copyright (C) 2014-2015 Red Hat Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *)

(* Implement 'mclu boot'. *)

module C = Libvirt.Connect
module D = Libvirt.Domain
module MS = Mclu_status

open Printf

open Utils

let memory = ref 0L                     (* 0 = choose for me *)
let set_memory s =
  try memory := bytes_of_human_size s
  with Not_found ->
    eprintf "mclu: don't understand --memory parameter '%s'
Try something like --memory 1G\n" s;
    exit 1
let size = ref 0L                       (* 0 = default *)
let set_size s =
  try size := bytes_of_human_size s
  with Not_found ->
    eprintf "mclu: don't understand --size parameter '%s'
Try something like --size 20G\n" s;
    exit 1
let timezone = ref ""                   (* "" = no timezone set *)
let vcpus = ref 0                       (* 0 = choose for me *)

let open_console = ref false
let open_viewer = ref false

let get_arg_speclist () = Arg.align [
  "--console",  Arg.Set open_console, " Open the serial console";
  "--cpus",     Arg.Set_int vcpus, "n Number of virtual CPUs";
  "--memory",   Arg.String set_memory, "nnG Amount of RAM to give guest";
  "--ram",      Arg.String set_memory, "nnG Amount of RAM to give guest";
  "--size",     Arg.String set_size, "nnG Size of disk to give guest";
  "--timezone", Arg.Set_string timezone, "TZ Set timezone of guest";
  "--vcpus",    Arg.Set_int vcpus, "n Number of virtual CPUs";
  "--viewer",   Arg.Set open_viewer, " Open the graphical console";
]

let boot ~verbose template name =
  let templates = Template.templates () in

  (* Does the template exist? *)
  let template_filename =
    try List.assoc template templates
    with Not_found ->
      eprintf "mclu: template %s not found
Try `mclu list --templates' to list all known templates.\n" template;
      exit 1 in

  (* Probe the template for various features. *)
  let template_info = Template.probe ~verbose template_filename in

  (* Check --size is not too small. *)
  let size =
    match !size, template_info.Template.minimum_size with
    | 0L, None -> 0L               (* virt-builder default *)
    | 0L, Some min_size ->         (* go with template minimum size *)
      min_size
    | size, Some min_size when size < min_size ->
      eprintf "mclu: --size parameter is smaller than the minimum specified by the template (%s).\n"
        (human_size min_size);
      exit 1
    | size, _ -> size in           (* go with user-specified size *)

  (* Decide how much RAM we will give the guest.  This affects our
   * choice of node, so do it early.
   *)
  let memory = !memory in
  let memory =
    if memory > 0L then (
      (* User requested, just check it's above the minimum. *)
      match template_info.Template.minimum_memory with
      | None -> memory
      | Some min when min > memory ->
        eprintf "mclu: minimum memory for this template is %s\n"
          (human_size min);
        exit 1
      | Some _ -> memory
    ) else (
      (* User didn't request any memory setting, use the recommended. *)
      match template_info.Template.recommended_memory with
      | Some memory -> memory
      | None -> 4L *^ 1024L *^ 1024L *^ 1024L (* 4 GB *)
    ) in

  (* Check what's running. *)
  let summary = MS.node_guest_summary ~verbose () in

  (* Did the user request a specific host?  If not, choose one. *)
  let hostname, name =
    match name_parse name with
    | Some hostname, name -> hostname, name
    | None, name ->
      (* Choose the first host with enough free memory. *)
      let nodes = List.filter (
        fun { MS.free_memory = free_memory } -> free_memory >= memory
      ) summary in
      match nodes with
      | [] ->
        eprintf "mclu: no node with enough free memory found
Try: `mclu status' and `mclu on <node>'\n";
        exit 1
      | node :: _ ->
        let hostname =
          node.MS.node_status.MS.node.Mclu_conf.hostname in
        hostname, name in

  (* Check there isn't a guest with this name running anywhere
   * in the cluster already.
   *)
  List.iter (
    fun ({ MS.active_guests = guests } as node) ->
      List.iter (
        fun { Mclu_list.dom_name = n } ->
          if name = n then (
            let hostname =
              node.MS.node_status.MS.node.Mclu_conf.hostname
            in
            eprintf "mclu: there is already a guest called '%s' (running on %s)\n"
              name hostname;
            exit 1
          )
      ) guests
  ) summary;

  (* Convert hostname to a specific node, and check it is up. *)
  let node =
    try List.find (
      fun node ->
        node.MS.node_status.MS.node.Mclu_conf.hostname = hostname
    ) summary
    with Not_found ->
      eprintf "mclu: no node is called '%s'\n" hostname;
      exit 1 in
  if not node.MS.node_status.MS.node_on then (
    eprintf "mclu: node '%s' is switched off
Try: `mclu on %s'\n" hostname hostname;
    exit 1
  );

  (* Where we upload the template and image on remote. *)
  let format, extension = "qcow2", "qcow2" in
  let remote_template = sprintf "/tmp/mclu%s.sh" (string_random8 ()) in
  let remote_template_wrapper = sprintf "/tmp/mclu%s.sh" (string_random8 ()) in
  let xml_template_wrapper = sprintf "/tmp/mclu%s.sh" (string_random8 ()) in
  let remote_image = sprintf "/var/tmp/%s.%s" name extension in
  let remote_external_kernel_dir = sprintf "/var/tmp/%s.boot" name in
  let remote_external_kernel = sprintf "/var/tmp/%s.boot/kernel" name in
  let remote_external_initrd = sprintf "/var/tmp/%s.boot/initrd" name in
  let remote_arch = node.MS.node_status.node_info.model in

  (* Guest arch defaults to the node host arch, but can be overridden
   * in the template.
   *)
  let guest_arch =
    match template_info.Template.guest_arch with
    | Some arch -> arch
    | None -> remote_arch in

  (* UEFI firmware and NVRAM on remote, if required. *)
  let nvram =
    match guest_arch with
    | "aarch64" ->
       Some ("/usr/share/edk2.git/aarch64/QEMU_EFI-pflash.raw",
             "/usr/share/edk2.git/aarch64/vars-template-pflash.raw",
             remote_image ^ ".nvram")
    | _ -> None in

  (* Get ready to generate the guest XML. *)
  let vcpus = !vcpus in
  let vcpus =
    if vcpus > 0 then vcpus
    else min 4 node.MS.node_status.MS.node_info.C.cpus in
  let mac_addr =
    sprintf "52:54:00:%02x:%02x:%02x"
      (Random.int 256) (Random.int 256) (Random.int 256) in

  (* Generate the guest XML. *)
  let generate_standard_xml () =
    (* XXX Better quoting. *)
    let xml = sprintf "\
<domain type='kvm'>
  <name>%s</name>
  <memory unit='KiB'>%Ld</memory>
  <currentMemory unit='KiB'>%Ld</currentMemory>
  <vcpu>%d</vcpu>
" name (memory /^ 1024L) (memory /^ 1024L) vcpus in

    let xml = xml ^ "\
  <os>
    <boot dev='hd'/>
" in
    let xml =
      match guest_arch with
      | "arm" | "armv7" | "armv7l" | "armv7hl" ->
         xml ^ "\
    <type arch='armv7l' machine='virt'>hvm</type>
"
      | "aarch64" ->
         xml ^ "\
    <type machine='virt'>hvm</type>
"
      | _ ->
         xml ^ "\
    <type>hvm</type>
" in

    let xml =
      match nvram with
      | Some (loader, nvram_template, nvram) ->
         xml ^ sprintf "\
    <loader readonly='yes' type='pflash'>%s</loader>
    <nvram template='%s'>%s</nvram>
" loader nvram_template nvram
      | None -> xml in

    let xml = xml ^
      if template_info.Template.needs_external_kernel then
        sprintf "\
    <kernel>%s</kernel>
    <initrd>%s</initrd>
" remote_external_kernel remote_external_initrd
      else "" in

    let xml = xml ^
      match template_info.Template.cmdline with
      | Some cmdline -> sprintf "    <cmdline>%s</cmdline>\n" cmdline
      | None -> "" in

    let xml = xml ^ "\
  </os>
  <features>
    <acpi/>
    <apic/>
    <pae/>
  </features>
  <cpu mode='host-passthrough'/> <!-- -cpu host, also allows nested -->
  <clock offset='utc'>
    <timer name='rtc' tickpolicy='catchup'/>
    <timer name='pit' tickpolicy='delay'/>
    <timer name='hpet' present='no'/>
  </clock>
  <on_poweroff>destroy</on_poweroff>
  <on_reboot>restart</on_reboot>
  <on_crash>restart</on_crash>
  <devices>
" in

    let xml = xml ^ sprintf "\
  <disk type='file' device='disk'>
    <driver name='qemu' type='%s' cache='none' io='native'/>
    <source file='%s'/>
" format remote_image in
    let xml = xml ^
    match template_info.Template.disk_bus with
    | Some "ide" ->
      "      <target dev='sda' bus='ide'/>\n"
    | Some "virtio" ->
      "      <target dev='vda' bus='virtio'/>\n"
    | Some "virtio-scsi" | None ->
      "      <target dev='sda' bus='scsi'/>\n"
    | Some bus ->
      eprintf "mclu: unknown disk-bus: %s\n" bus;
      exit 1 in
    let xml = xml ^ "\
    </disk>
" in

    let xml =
      xml ^
        if template_info.Template.disk_bus = Some "virtio-scsi" then
          "  <controller type='scsi' index='0' model='virtio-scsi'/>\n"
        else
          "" in

    (* XXX Don't hard-code bridge name here. *)
    let network_model =
      match template_info with
      | { Template.network_model = None } -> "virtio"
      | { Template.network_model = Some d } -> d in
    let xml = xml ^ sprintf "\
    <interface type='bridge'>
      <mac address='%s'/>
      <source bridge='br0'/>
      <model type='%s'/>
    </interface>
" mac_addr network_model in

    let xml = xml ^ "\
    <serial type='pty'>
      <target port='0'/>
    </serial>
    <console type='pty'>
      <target type='serial' port='0'/>
    </console>
" in
    let xml =
      match guest_arch with
      | "i386" | "i486" | "i586" | "i686"
      | "x86_64" ->
         xml ^ "\
    <input type='tablet' bus='usb'/>
    <input type='mouse' bus='ps2'/>
    <input type='keyboard' bus='ps2'/>
    <graphics type='vnc' autoport='yes'/>
    <video>
      <model type='cirrus' vram='9216' heads='1'/>
    </video>
"
      | _ -> xml in
    let xml = xml ^ "\
  </devices>
</domain>" in
    xml

  and generate_custom_xml () =
    (* Generate a wrapper script to make passing the variables
     * to the template easier.
     *)
    let () =
      let chan = open_out xml_template_wrapper in
      let fpf fs = fprintf chan fs in
      fpf "#!/bin/sh\n";
      fpf "export format=%s\n" (quote format);
      fpf "export initrd=%s\n" (quote remote_external_initrd);
      fpf "export kernel=%s\n" (quote remote_external_kernel);
      fpf "export mac_addr=%s\n" (quote mac_addr);
      fpf "export memory_kb=%Ld\n" (memory /^ 1024L);
      fpf "export name=%s\n" (quote name);
      fpf "export output=%s\n" (quote remote_image);
      fpf "export vcpus=%d\n" vcpus;
      fpf "%s xml\n" template_filename;
      close_out chan;
      Unix.chmod xml_template_wrapper 0o755 in

    if verbose then printf "%s\n%!" xml_template_wrapper;
    let chan = Unix.open_process_in xml_template_wrapper in
    let lines = ref [] in
    (try while true do lines := input_line chan :: !lines done
     with End_of_file -> ());
    let stat = Unix.close_process_in chan in
    (match stat with
     | Unix.WEXITED 0 -> ()
     | Unix.WEXITED i ->
        eprintf "mclu: template '%s' subcmd xml exited with error %d\n"
                template_filename i;
        exit 1
     | Unix.WSIGNALED i ->
        eprintf "mclu: template '%s' subcmd xml killed by signal %d\n"
                template_filename i;
        exit 1
     | Unix.WSTOPPED i ->
        eprintf "mclu: template '%s' subcmd xml stopped by signal %d\n"
                template_filename i;
        exit 1
    );
    let xml = String.concat "\n" (List.rev !lines) in
    xml
  in

  let xml =
    if not template_info.Template.has_xml_target then
      generate_standard_xml ()
    else
      generate_custom_xml () in

  (* Copy the template to remote. *)
  let cmd =
    sprintf "scp %s root@%s:%s"
      (quote template_filename) (quote hostname) remote_template in
  if verbose then printf "%s\n%!" cmd;
  if Sys.command cmd <> 0 then (
    eprintf "mclu: scp template to remote failed\n";
    exit 1
  );

  (* Create a wrapper script that sets the variables and runs the
   * template.  This just avoids complex quoting.
   *)
  let () =
    let chan = open_out remote_template_wrapper in
    let fpf fs = fprintf chan fs in
    fpf "#!/bin/bash\n";
    fpf "set -e\n";
    (* XXX Don't hard-code network_bridge here. *)
    fpf "export LIBGUESTFS_BACKEND_SETTINGS=network_bridge=br0\n";
    fpf "export base_image=%s\n" (quote template_info.Template.base_image);
    fpf "export format=%s\n" (quote format);
    fpf "export guest_arch=%s\n" (quote guest_arch);
    fpf "export name=%s\n" (quote name);
    fpf "export output=%s\n" (quote remote_image);
    (match size with
    | 0L -> ()
    | size -> fpf "export size=%s\n" (quote (sprintf "--size %Ldb" size))
    );
    (match !timezone with
    | "" -> ()
    | tz -> fpf "export timezone=%s\n" (quote (sprintf "--timezone %s" tz))
    );
    (match nvram with
     | Some (_, nvram_template, nvram) ->
        fpf "cp %s %s\n" (quote nvram_template) (quote nvram)
     | None -> ()
    );
    fpf "%s build\n" remote_template;
    if template_info.Template.needs_external_kernel then (
      fpf "rm -rf %s\n" (quote remote_external_kernel_dir);
      fpf "mkdir %s\n" (quote remote_external_kernel_dir);
      fpf "pushd %s\n" (quote remote_external_kernel_dir);
      fpf "virt-builder --get-kernel %s\n" (quote remote_image);
      fpf "ln vmlinuz-* kernel\n";
      fpf "ln init* initrd\n";
      fpf "popd\n";
    );
    close_out chan;
    Unix.chmod remote_template_wrapper 0o755 in

  let cmd =
    sprintf "scp %s root@%s:%s"
      (quote remote_template_wrapper) (quote hostname)
      (quote remote_template_wrapper) in
  if verbose then printf "%s\n%!" cmd;
  if Sys.command cmd <> 0 then (
    eprintf "mclu: scp template wrapper to remote failed\n";
    exit 1
  );

  let cmd =
    sprintf "ssh root@%s %s" (quote hostname) (quote remote_template_wrapper) in
  if verbose then printf "%s\n%!" cmd;
  if Sys.command cmd <> 0 then (
    eprintf "mclu: remote build failed\n";
    exit 1
  );

  (* Start the guest. *)
  let dom =
    try
      let conn =
        let name = node.MS.node_status.MS.node.Mclu_conf.libvirt_uri in
        C.connect ~name () in
      let dom = D.create_xml conn xml [] in
      printf "mclu: %s:%s started\n" hostname (D.get_name dom);
      dom
    with Libvirt.Virterror msg ->
      eprintf "mclu: %s: %s\n" hostname (Libvirt.Virterror.to_string msg);
      exit 1 in

  (* Graphical console? *)
  if !open_viewer then
    Mclu_viewer.viewer ~verbose ~host:hostname (D.get_name dom);

  (* Serial console?  (Interactive, so run it last) *)
  if !open_console then
    Mclu_console.console ~verbose ~host:hostname (D.get_name dom)

let run ~verbose = function
  | [ template; name ] ->
    boot ~verbose template name
  | _ ->
    eprintf "Usage: mclu boot <template> <[host:]name>\n";
    exit 1
